/*
 * 
 */

#include "OperatorInterface.h"
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#define OLED_RESET 12
Adafruit_SSD1306 display(OLED_RESET);

#define OP_ID 50

OperatorInterface opinterface;

void buttonChanged(int button, int val);
void modeChanged(int mode);
void robotAdded();
void robotRemoved(bool connection);

void receivedAction(Msg m);
void commsTimeout();
void controllerAdded();
void controllerRemoved();

void setup() {
  delay(2000);
  Serial.begin(9600);

  display.begin(SSD1306_SWITCHCAPVCC, 0x3D);  // initialize with the I2C addr 0x3D (for the 128x64)
  display.clearDisplay();
  displayLogo();
  
  opinterface = OperatorInterface();

  opinterface.begin();
  opinterface.setOpID(OP_ID);
  opinterface.initOperator(XBEE_TRANSPARENT_CONN, 9600, &Serial1);

 /*
  // this is if we want to take a photo haha
  for(int i=0; i<6; i++) {
    opinterface.ledQuickFade(opinterface.led_pins[i], 0, 255);
  }
  while(1<3) {
    displayLogo();
  }
  */

  //opinterface.TESTING = true;
  opinterface.setAutoconnect(true);
  opinterface.set_received_action_callback(receivedAction);
  opinterface.set_comms_timeout_callback(commsTimeout);
  opinterface.set_controller_added_callback(controllerAdded);
  opinterface.set_controller_removed_callback(controllerRemoved);
  
  opinterface.set_button_changed_callback(buttonChanged);
  opinterface.set_mode_changed_callback(modeChanged);
  opinterface.set_robot_added_callback(robotAdded);
  opinterface.set_robot_removed_callback(robotRemoved);

  // 1. control
  // drive, arm, collect, 
  // o-c, deposit, wave
  setButtonLabel("Drive", 0, 1);
  setButtonLabel("Arm", 1, 1);
  setButtonLabel("Auto", 2, 1);
  setButtonLabel("O-C", 3, 1);
  setButtonLabel("Grab", 4, 1);
  setButtonLabel("Wave", 5, 1);

  // 2. precision combo mode
  // l, fwd, r, 
  // +, bwd, -
  setButtonLabel("L", 0, 2);
  setButtonLabel("Fwd", 1, 2);
  setButtonLabel("R", 2, 2);
  setButtonLabel("+", 3, 2);
  setButtonLabel("Bwd", 4, 2);
  setButtonLabel("-", 5, 2);

  // 3. claw
  // 1 - lower, 3 - hold, 4 - deposit
  // 2 - small, 2 - med, 2 - large
  setButtonLabel("1.Low", 0, 3);
  setButtonLabel("3.Hold", 1, 3);
  setButtonLabel("4.Dep", 2, 3);
  setButtonLabel("2.Smll", 3, 3);
  setButtonLabel("2.Med", 4, 3);
  setButtonLabel("2.Lrg", 5, 3);

  setModeLabel("Control", 1);
  setModeLabel("Precision", 2);
  setModeLabel("Claw", 3);

}

void loop() {

  opinterface.updateOperator();

  if(opinterface.isConnectedToRobot()) {
    mainMenu(opinterface.button_states, opinterface.CURRENT_MODE);
  } else {
    displaySearching(millis());
  }

  if(opinterface.isConnectedToRobot()) {
  
    if(opinterface.getCurrentMode() == 1) {

      // control
      // drive, arm, blink, wave
    
      if(opinterface.getButton(0) == 1) { // drive state
        opinterface.joystickDriveControl(255, false);
      }
  
      if(opinterface.getButton(1) == 1) { // arm state
        opinterface.joystickArmControl();
      }

      for(int i=2; i<6; i++) {
        if(opinterface.getButton(i) == 1) {
          opinterface.setButtonState(i, 0);
        }
      }

    } else if(opinterface.getCurrentMode() == 2) {

      // for precision combo mode, all of the commands
      // are momentary, so we can switch off all the leds.
      for(int i=0; i<6; i++) {
        if(opinterface.getButton(i) == 1) {
          opinterface.setButtonState(i, 0);
        }
      }
      
    } else if(opinterface.getCurrentMode() == 3) {

      // for claw mode, all of the commands are also
      // momentary - they aren't continuing sequences.
      // so we can switch off all the leds.

      for(int i=0; i<6; i++) {
        if(opinterface.getButton(i) == 1) {
          opinterface.setButtonState(i, 0);
        }
      }

    }

  }
  
}

void buttonChanged(int button, int val) {
  // This is called each time a button is pressed
}

void modeChanged(int mode) {
  // This is called each time the mode switch is changed
}

void robotAdded() {
  // A new robot has been added to the list to select from.
  // If you want to select it, you could call 
  // opinterface.xbeeChooseRobotToConnect(),
  // but make sure that SELECTED_ROBOT is false first.
}

void robotRemoved(bool connection) {
  // A robot has been removed from the list to select from.
  // If connection == true, we are still connected to the robot we're
  // controlling. If not, you might want to prompt the user to select
  // a new robot. See comment in robotAdded() for details.
}

void receivedAction(Msg m) {
  // Received an action from the controller. The data is
  // packed into the Msg struct. Core actions with this data
  // will be done inside of the main robot mission program.
  // All Msgs will be passed through to this function, even
  // if there is / is not a core action associated with it.
  // You can do custom actions with this data here.

  /*
  Serial << "---RECEIVED ACTION---" << endl;
  Serial << "action: " << m.action << endl;
  Serial << "command: " << m.pck1.cmd << endl;
  Serial << "key: " << m.pck1.key << endl;
  Serial << "val: " << m.pck1.val << endl;
  Serial << "command: " << m.pck2.cmd << endl;
  Serial << "key: " << m.pck2.key << endl;
  Serial << "val: " << m.pck2.val << endl;
  Serial << "delim: " << m.delim << endl;
  */
  
}

void commsTimeout() {
  // The comms timed out. You can do an action here, such as
  // turning off the motors and sending them back to the 
  // home positions.
}

void controllerAdded() {
  // Called when receiving an Xbee response. The ID of the
  // controller will be sent via the received action. You
  // could do an action here, such as prepare the robot's
  // servos for moving.
}

void controllerRemoved() {
  // Called when the Xbee watchdog detects no messages
  // received from a controller after a given amount of time.
  // The ID of the controller could be deduced by not hearing
  // from it in the received action. You could do an action
  // here, such as the robot waving goodbye.
}
