void comms_navpi() {
  
  current_time = millis();

  navcam_comms.updateComms();

  if(current_time-last_latency_print >= 500) {
    Serial << "Comm latency: " << navcam_comms.getCommLatency() << endl;
    last_latency_print = current_time;
  }

  // send update every 20 ms
  if(current_time-last_update >= 20) {
    Serial << "!!!!!! Navcam sending sensor data" << endl;
    //Serial1.println("#L1,133,R1,134!");
    //delay(500);
    
    if(blink456) {
      bowie.bowielights.setLight(99, MIN_BRIGHTNESS);
    } else {
      bowie.bowielights.setLight(99, 0);
    }
    blink456 = !blink456;
    
    comms_navpi_send_sensors();
    last_update = current_time;
  }

  // if it hasn't received anything in a while, kick it to send a message
  if(current_time-last_update >= 1000 && current_time-navpi_last_rx >= 1000) {
    //Serial << "!!!!!!!!!!!!!!!!!! navcam sending message !!!!!!!!!!!!!!!!!" << endl;
    if(flip_coin == true) {
      navcam_comms.connSend('#', 'J', 1, 0, 'Q', 2, 0, '!');
      flip_coin = !flip_coin;
    } else {
      navcam_comms.connSend('#', 'J', 3, 0, 'Q', 4, 0, '!');
      flip_coin = !flip_coin;
    }
    last_update = current_time;
  }
  
}

// sending all the sensors
void comms_navpi_send_sensors() {

  switch(sensor_send_index) {
    case 0: { // current sensors
      bowie.motorCurrent.updateCurrentSensor(); 
      bowie.servoCurrent.updateCurrentSensor();
      uint16_t val1 = (uint16_t)bowie.motorCurrent.getCurrentSensorReading();
      uint16_t val2 = (uint16_t)bowie.servoCurrent.getCurrentSensorReading();
      navcam_comms.connSend('#', 'E', 1, val1, 'F', 1, val2, '!');
    }
    break;
    case 1: { // motor dir and speed
      uint8_t key1 = 99;
      if(bowie.bowiedrive.getMotorADir() == true) {
        key1 = 1;
      } else {
        key1 = 0;
      }
      uint16_t val1 = (uint16_t)bowie.bowiedrive.getMotorASpeed();
      uint8_t key2 = 99;
      if(bowie.bowiedrive.getMotorBDir() == true) {
        key2 = 1;
      } else {
        key2 = 0;
      }
      uint8_t val2 = (uint16_t)bowie.bowiedrive.getMotorBSpeed();
      navcam_comms.connSend('#', 'H', key1, val1, 'I', key2, val2, '!');
    }
    break;
    case 2: { // servos - arm l & r
      uint16_t val1 = (uint16_t)bowie.bowiearm.getArmPos();
      uint16_t val2 = (uint16_t)bowie.bowiearm.getArmPos();
      navcam_comms.connSend('#', 'A', 1, val1, 'A', 2, val2, '!');
    }
    break;
    case 3: { // servos - end & hopper
      uint16_t val1 = (uint16_t)bowie.bowiescoop.getEndPos();
      uint16_t val2 = (uint16_t)bowie.bowiehopper.getHopperPos();
      navcam_comms.connSend('#', 'A', 3, val1, 'A', 4, val2, '!');
    }
    break;
    case 4: { // servos - lid & extra
      uint16_t val1 = (uint16_t)bowie.bowiehopper.getLidPos();
      uint16_t val2 = (uint16_t)bowieclaw.getClawPos();
      navcam_comms.connSend('#', 'A', 5, val1, 'A', 6, val2, '!');
    }
    break;
    case 5: { // leds - front left & right
      uint16_t val1 = (uint16_t)bowie.bowielights.getLight(0); // front left
      uint16_t val2 = (uint16_t)bowie.bowielights.getLight(1); // front right
      navcam_comms.connSend('#', 'D', 1, val1, 'D', 2, val2, '!');
    }
    break;
    case 6: { // leds - back left & right
      uint16_t val1 = (uint16_t)bowie.bowielights.getLight(2); // back left
      uint16_t val2 = (uint16_t)bowie.bowielights.getLight(3); // back right
      navcam_comms.connSend('#', 'D', 3, val1, 'D', 4, val2, '!');
    }
    break;
    case 7: { // humidity and temperature
      int chk = DHT.read22(DHT22_PIN);
      uint16_t val1 = DHT.humidity;
      uint16_t val2 = DHT.temperature;
      navcam_comms.connSend('#', 'K', 1, val1, 'K', 2, val2, '!'); 
    }
    break;
    case 8: { // battery
      float batt_val = ((float)analogRead(A19)/1024.0*3.3*2.76);
      batt_val = batt_val*100;
      uint16_t val1 = (uint16_t)batt_val;
      uint16_t val2 = (uint16_t)batt_val;
      navcam_comms.connSend('#', 'O', 1, val1, 'O', 2, val2, '!');
    }
    break;
    case 9: { // gps latitude
      uint8_t key1 = 99;
      if(gps.location.lat() >= 0) {
        key1 = 1;
      } else {
        key1 = 0;
      }
      int integer_part = (int)abs(gps.location.lat());
      int decimal_part = (int)(1000.0 * (abs(gps.location.lat()) - (float)integer_part));
      uint16_t val1 = (uint16_t)integer_part;
      uint16_t val2 = (uint16_t)decimal_part;
      navcam_comms.connSend('#', 'T', key1, val1, 'T', 9, val2, '!');
    }
    break;
    case 10: { // gps longitude
      uint8_t key1 = 99;
      if(gps.location.lng() >= 0) {
        key1 = 1;
      } else {
        key1 = 0;
      }
      int integer_part = (int)abs(gps.location.lng());
      int decimal_part = (int)(1000.0 * (abs(gps.location.lng()) - (float)integer_part));
      uint16_t val1 = (uint16_t)integer_part;
      uint16_t val2 = (uint16_t)decimal_part;
      navcam_comms.connSend('#', 'U', key1, val1, 'U', 9, val2, '!');
    }
    break;
    case 11: { // compass heading
      uint8_t key1 = 99;
      if(compass_heading >= 0) {
        key1 = 1;
      } else {
        key1 = 0;
      }
      int integer_part = (int)compass_heading;
      int decimal_part = (int)(1000.0 * (compass_heading - (float)integer_part));
      uint16_t val1 = (uint16_t)integer_part;
      uint16_t val2 = (uint16_t)decimal_part;
      navcam_comms.connSend('#', 'V', key1, val1, 'V', 9, val2, '!');
    }
    break;
    case 12: { // altitude
      uint8_t key1 = 99;
      if(gps.altitude.meters() >= 0) {
        key1 = 1;
      } else {
        key1 = 0;
      }
      int integer_part = (int)gps.altitude.meters();
      int decimal_part = (int)(1000.0 * (gps.altitude.meters() - (float)integer_part));
      uint16_t val1 = (uint16_t)integer_part;
      uint16_t val2 = (uint16_t)decimal_part;
      navcam_comms.connSend('#', 'a', key1, val1, 'a', 9, val2, '!');
    }
    break;
    case 13: { // imu pitch
      uint8_t key1 = 99;
      if(compass_pitch >= 0) {
        key1 = 1;
      } else {
        key1 = 0;
      }
      int integer_part = (int)abs(compass_pitch);
      int decimal_part = (int)(1000.0 * (abs(compass_pitch) - (float)integer_part));
      uint16_t val1 = (uint16_t)integer_part;
      uint16_t val2 = (uint16_t)decimal_part;
      navcam_comms.connSend('#', 'b', key1, val1, 'b', 9, val2, '!');
    }
    break;
    case 14: { // imu roll
      uint8_t key1 = 99;
      if(compass_roll >= 0) {
        key1 = 1;
      } else {
        key1 = 0;
      }
      int integer_part = (int)abs(compass_roll);
      int decimal_part = (int)(1000.0 * (abs(compass_roll) - (float)integer_part));
      uint16_t val1 = (uint16_t)integer_part;
      uint16_t val2 = (uint16_t)decimal_part;
      navcam_comms.connSend('#', 'c', key1, val1, 'c', 9, val2, '!');
    }
    break;
    case 15: { // imu yaw
      uint8_t key1 = 0;
      uint16_t val1 = 0;
      uint16_t val2 = 0;
      navcam_comms.connSend('#', 'd', key1, val1, 'd', 9, val2, '!');
    }
    break;
    case 16: { // gps sats & hdop
      uint16_t val1 = (uint16_t)gps.satellites.value();
      uint16_t val2 = (uint16_t)floor(gps.hdop.value());
      navcam_comms.connSend('#', 'e', 1, val1, 'e', 2, val2, '!');
    }
    break;
  }
  sensor_send_index++;

  if(sensor_send_index > 16) {
    sensor_send_index = 0;
  }
  
}

void receivedAction(Msg m) {

  // Received an action from the controller. The data is
  // packed into the Msg struct. Core actions with this data
  // will be done inside of the main robot mission program.
  // All Msgs will be passed through to this function, even
  // if there is / is not a core action associated with it.
  // You can do custom actions with this data here.

  Serial << "---NAVPI RECEIVED ACTION---" << endl;
  Serial << "action: " << m.action << endl;
  Serial << "command: " << m.pck1.cmd << endl;
  Serial << "key: " << m.pck1.key << endl;
  Serial << "val: " << m.pck1.val << endl;
  Serial << "command: " << m.pck2.cmd << endl;
  Serial << "key: " << m.pck2.key << endl;
  Serial << "val: " << m.pck2.val << endl;
  Serial << "delim: " << m.delim << endl;

  navpi_last_rx = millis();

  // this is essentially the same as control(Msg m)
  
  Packet packets[2] = { m.pck1, m.pck2 };

  bowie.bowielogger.logEvent(now(), m.action, m.pck1.cmd, m.pck1.key, m.pck1.val, m.pck2.cmd, m.pck2.key, m.pck2.val, m.delim);

  bowie.joystickDriveControl(m);
  bowie.joystickArmControl(m);

  // 1. control
  controlMode(m, packets);
  
  // 2. precision drive mode
  precisionDriveMode(m, packets);

  // 3. all the commands for claw mode
  clawMode(m, packets);

  // (not in use) 2. precision arm mode
  // precisionArmMode();
  
}

void commsTimeout() {
  // The comms timed out. You can do an action here, such as
  // turning off the motors and sending them back to the 
  // home positions.
}

void controllerAdded() {
  // Called when receiving an Xbee response. The ID of the
  // controller will be sent via the received action. You
  // could do an action here, such as prepare the robot's
  // servos for moving.
}

void controllerRemoved() {
  // Called when the Xbee watchdog detects no messages
  // received from a controller after a given amount of time.
  // The ID of the controller could be deduced by not hearing
  // from it in the received action. You could do an action
  // here, such as the robot waving goodbye.
}

void updateRobotsPeriodicMessages() {
  // Updating our periodic messages

  random_periodic1.priority = 10;
  random_periodic1.action = '@';
  random_periodic1.pck1.cmd = 'Z';
  random_periodic1.pck1.key = 1;
  random_periodic1.pck1.val = (int)random(0, 1024);
  random_periodic1.pck2.cmd = 'Z';
  random_periodic1.pck2.key = 1;
  random_periodic1.pck2.val = (int)random(0, 1024);
  random_periodic1.delim = '!';

  random_periodic2.priority = 9;
  random_periodic2.action = '@';
  random_periodic2.pck1.cmd = 'Z';
  random_periodic2.pck1.key = 1;
  random_periodic2.pck1.val = (int)random(1000, 1024);
  random_periodic2.pck2.cmd = 'Z';
  random_periodic2.pck2.key = 1;
  random_periodic2.pck2.val = (int)random(1000, 1024);
  random_periodic2.delim = '!';

  navcam_comms.updatePeriodicMessage(random_periodic1);
  navcam_comms.updatePeriodicMessage(random_periodic2);
  
}
