void compassInit() {

  if (!bno.begin()) {
    Serial.println("Oops, no BNO055 detected ... Check your wiring or I2C ADDR!");
    COMPASS_ENABLED = false;
    loadingIndicator(5, 255, 0, 0); // red
    
    for(int i=0; i<3; i++) { // retry 3 times
      unloadingIndicator(5, 10, 0, 0);
      delay(1000);
      compassInit();
      delay(500);
      if (!bno.begin()) {
        Serial.println("Oops, no BNO055 detected ... Check your wiring or I2C ADDR!");
        COMPASS_ENABLED = false;
        loadingIndicator(5, 255, 0, 0); // red
      } else {
        break;
      }
    }
    
  }

  if(COMPASS_ENABLED) {
    compassLoadCalibration();
    
    bno.setExtCrystalUse(true);
    delay(20);
    loadingIndicator(5, 0, 255, 0); // green
    delay(1000);
    loadingIndicator(5, 0, 0, 0); // green
  }

  emptyPixels();
  
}

void compassUpdate() {

  if(!COMPASS_ENABLED) return;
  
  sensors_event_t event;
  bno.getEvent(&event);

  compass_heading = event.orientation.x;
  compass_roll = event.orientation.y;
  compass_pitch = event.orientation.z;

  // http://www.geomag.nrcan.gc.ca/calc/mdcal-r-en.php?date=2018-08-30&latitude=45.394794&latitude_direction=1&longitude=75.761199&longitude_direction=-1
  float rd = 13.0; // declination
  float ro = 0.0;//45.0; // offset
  float rx = event.orientation.x; // x
  float rh = (270-(rx+rd))+270; // heading
  if(rh > 360.0) {
    rh -= 360;
  }
  rh -= ro;
  compass_heading = rh;

  delay(BNO055_SAMPLERATE_DELAY_MS);

}

int distanceAngles(int alpha, int beta) {
  int phi = abs(beta - alpha) % 360;       // This is either the distance or 360 - distance
  int distance = phi > 180 ? 360 - phi : phi;
  //calculate sign 
  int sign = (alpha - beta >= 0 && alpha - beta <= 180) || (alpha - beta <= -180 && alpha - beta >= -360) ? 1 : -1; 
  distance *= sign;
  return distance;
}

void compassLoadCalibration() {
  eeAddress = BNO_EEPROM_ADDR;

  Serial.println("\nFound Calibration for this sensor in EEPROM.");
  eeAddress += sizeof(long);
  EEPROM.get(eeAddress, calibrationData);

  displaySensorOffsets(calibrationData);

  Serial.println("\n\nRestoring Calibration data to the BNO055...");
  bno.setSensorOffsets(calibrationData);

  Serial.println("\n\nCalibration data loaded into BNO055");
  
  Serial.println("\n--------------------------------\n");
}

void displaySensorOffsets(const adafruit_bno055_offsets_t &calibData) {
    Serial.print("Accelerometer: ");
    Serial.print(calibData.accel_offset_x); Serial.print(" ");
    Serial.print(calibData.accel_offset_y); Serial.print(" ");
    Serial.print(calibData.accel_offset_z); Serial.print(" ");

    Serial.print("\nGyro: ");
    Serial.print(calibData.gyro_offset_x); Serial.print(" ");
    Serial.print(calibData.gyro_offset_y); Serial.print(" ");
    Serial.print(calibData.gyro_offset_z); Serial.print(" ");

    Serial.print("\nMag: ");
    Serial.print(calibData.mag_offset_x); Serial.print(" ");
    Serial.print(calibData.mag_offset_y); Serial.print(" ");
    Serial.print(calibData.mag_offset_z); Serial.print(" ");

    Serial.print("\nAccel Radius: ");
    Serial.print(calibData.accel_radius);

    Serial.print("\nMag Radius: ");
    Serial.print(calibData.mag_radius);
}
