void ledsInit() {
  
  for(int i=0; i<NUMPIXELS; i++) {
    pixel_colours[i] = 0;
    pixel_states[i] = 0; // 0 = off, 1 = indicators, 2 = turning
  }
  
  pixels.begin();
  
}

void updateLeds() {

  // now update the pixels
  for(int i=0; i<NUMPIXELS; i++) {
    if(pixel_states[i] > 0) {
      pixels.setPixelColor(i, pixel_colours[i]);
    } else {
      pixels.setPixelColor(i, pixels.Color(0, 0, 0));
    }
  }
  pixels.show();
  emptyPixels();

}

void loadingIndicator(uint8_t wait, uint8_t r, uint8_t g, uint8_t b) {
  for(int i=0; i<pixels.numPixels(); i++){
    pixels.setPixelColor(i, pixels.Color(r,g,b));
    pixels.show();
    delay(wait);
  }
}

void unloadingIndicator(uint8_t wait, uint8_t r, uint8_t g, uint8_t b) {
  for(int i=pixels.numPixels(); i>=0; i--) {
    pixels.setPixelColor(i, pixels.Color(r,g,b));
    pixels.show();
    delay(wait);
  }
}

void ledsOn(int alt, uint8_t r, uint8_t g, uint8_t b) {
  for(int i=0; i<pixels.numPixels(); i++) {
    if(alt == 0) {
      if(i%2 == 0) {
        pixels.setPixelColor(i, pixels.Color(r,g,b));
      } else {
        pixels.setPixelColor(i, pixels.Color(0,0,0));
      }
    } else if(alt == 1) {
      if(i%2 == 0) {
        pixels.setPixelColor(i, pixels.Color(0,0,0));
      } else {
        pixels.setPixelColor(i, pixels.Color(r,g,b));
      }
    } else {
      pixels.setPixelColor(i, pixels.Color(r,g,b));
    }
  }
  pixels.show();
}

void ledsOff() {
  for(int i=0; i<pixels.numPixels(); i++) {
    pixels.setPixelColor(i, pixels.Color(0,0,0));
  }
  pixels.show();
}

// layer one of the leds - the compass indicators
// north = green
// robot = red
void compassIndicatorLeds() {

  for(int i=NUMPIXELS; i>=0; i--) {
    pixels.setPixelColor(i, pixels.Color(0,0,0));
  }

  // the neopixel directly infront of the robot on the
  // compass viewer
  robot_direction_indicator = 11;//2;

  // divide compass heading by 15 to map to the 
  // number of leds (because 360/15 == 24).
  // then subtract 24, to get the led number to
  // be lit. 24-(360/15) == 0.
  //north_indicator = 24-floor(compass_heading/15);
  north_indicator = floor((360-compass_heading)/15);
  //north_indicator = floor((compass_heading)/15);
  north_indicator += robot_direction_indicator-1;
  if(north_indicator < 0) north_indicator = 0;
  if(north_indicator >= 24) north_indicator -= 24;

  // default tracking heading is 45
  /*
  heading_indicator = 0;
  heading_indicator = floor(tracking_heading/15);
  heading_indicator += north_indicator;
  heading_indicator += robot_direction_indicator-1;
  if(heading_indicator < 0) heading_indicator = 0;
  if(heading_indicator > 24) heading_indicator -= 24;
  */
  
  // north = green
  pixel_states[north_indicator] = 1;
  pixel_colours[north_indicator] = pixels.Color(0,255,0);
  
  // heading = green
  //pixel_states[heading_indicator] = 1;
  //pixel_colours[heading_indicator] = pixels.Color(0,255,0);
  
  // robot = red
  pixel_states[robot_direction_indicator] = 1;
  pixel_colours[robot_direction_indicator] = pixels.Color(255,0,0);
  
  // pixels will be updated in the main loop
  
}

// update the leds based on drive state
// this is layer two of the leds
void compassDriveLeds() {
  
  uint32_t compass_pixel_colour = pixels.Color(255, 255, 255);
  
  switch(DRIVE_STATE) {
    case IDLE_DRIVE:
    break;
    case FORWARD:
      // the span of leds we want to update
      for(int i=0; i<=4; i++) {
        // check to make sure these aren't being lit as an indicator
        if(pixel_states[i] != 1) {
          pixel_states[i] = 2;
          pixel_colours[i] = compass_pixel_colour;
        }
      }
    break;
    case RIGHT:
      for(int i=6; i<=10; i++) {
        if(pixel_states[i] != 1) {
          pixel_states[i] = 2;
          pixel_colours[i] = compass_pixel_colour;
        }
      }
    break;
    case HARD_RIGHT:
      for(int i=4; i<=12; i++) {
        if(pixel_states[i] != 1) {
          pixel_states[i] = 2;
          pixel_colours[i] = compass_pixel_colour;
        }
      }
    break;
    case LEFT:
      for(int i=18; i<=22; i++) {
        if(pixel_states[i] != 1) {
          pixel_states[i] = 2;
          pixel_colours[i] = compass_pixel_colour;
        }
      }
    break;
    case HARD_LEFT:
      for(int i=16; i<=23; i++) {
        if(pixel_states[i] != 1) {
          pixel_states[i] = 2;
          pixel_colours[i] = compass_pixel_colour;
        }
      }
      if(pixel_states[0] != 1) {
        pixel_states[0] = 2;
        pixel_colours[0] = compass_pixel_colour;
      }
    break;
    case REVERSE:
    break;
  }

}

// layer three of the leds - the background
void compassBgLeds(int alt, uint8_t r, uint8_t g, uint8_t b) {

  for(int i=0; i<pixels.numPixels(); i++) {
    if(alt == 0) {
      if(i%2 == 0) {
        if(pixel_states[i] == 0) {
          pixel_states[i] = 3;
          pixel_colours[i] = pixels.Color(r, g, b);
        }
      } else {
        if(pixel_states[i] == 0) {
          pixel_states[i] = 3;
          pixel_colours[i] = pixels.Color(0, 0, 0);
        }
      }
    } else if(alt == 1) {
      if(i%2 == 0) {
        if(pixel_states[i] == 0) {
          pixel_states[i] = 3;
          pixel_colours[i] = pixels.Color(0, 0, 0);
        }
      } else {
        if(pixel_states[i] == 0) {
          pixel_states[i] = 3;
          pixel_colours[i] = pixels.Color(r, g, b);
        }
      }
    } else {
      if(pixel_states[i] == 0) {
        pixel_states[i] = 3;
        pixel_colours[i] = pixels.Color(r, g, b);
      }
    }
  }

}

void emptyPixels() {
  for(int i=0; i<pixels.numPixels(); i++) {
    pixel_colours[i] = 0;
    pixel_states[i] = 0;
  }
  bg_pixel_colour = pixels.Color(0, 0, 0);
}
