/*
 * Green Board Bowie
 * 
 */
#include "BowieClaw.h"
#include "MegaBowieShoreline.h"
#include "BowieComms.h"

#define CONSOLE_DEBUG false

// Pi Comms
BowieComms navcam_comms = BowieComms();

// Neopixels for compass
#include <Adafruit_NeoPixel.h>

// GPS
#include <TinyGPS++.h>

// DHT22
#include <dht.h>

// Compass
#include <Adafruit_Sensor.h>
#include <Adafruit_BNO055.h>
#include <utility/imumaths.h>
#include <Wire.h>
#include <EEPROM.h>

#define ROBOT_ID 3

MegaBowieShoreline bowie;
long last_bowie_log = 0;

// ---- Compass Neopixel
/* connected to Sensor Input 1:
 * Gnd      -  Gnd (Grey)
 * 5V uC    -  Pwr (Red)
 * 3.3V ext -  NC
 * A21      -  NC
 * 39 / A20 -  Sig (White)
 */
#define PIN 54
#define NUMPIXELS 24
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800);

// ---- GPS
// connects to Serial3
/* Connected to Navigation slot:
 * Gnd - Black
 * 3.3V ext - White
 * Fix (28) - Blue
 * RST (29) - Orange
 * TX3 (8) - Grey
 * RX3 (7) - Red
 * SDA (18) - Green
 * SCL (19) - Yellow
 */
static const uint32_t GPSBaud = 9600;
TinyGPSPlus gps;
#define GPS_FIX_PIN 28
uint8_t printCount = 0;
#define MAX_POINTS 1
double POINTS_LAT[MAX_POINTS];
double POINTS_LON[MAX_POINTS];
uint8_t POINTS_ACTION[MAX_POINTS];
uint8_t current_point = 0;
#define DISTANCE_THRESH 3.1 // meters
bool following_course = true;

// ---- Compass
#define BNO055_SAMPLERATE_DELAY_MS (100)
#define IMU_RESET_PIN 57
Adafruit_BNO055 bno = Adafruit_BNO055();
bool COMPASS_ENABLED = true;
int LED_BRIGHTNESS = 255;
adafruit_bno055_offsets_t calibrationData;
sensor_t sensor;
#define BNO_EEPROM_ADDR 512
int eeAddress = BNO_EEPROM_ADDR;
double compass_heading = 0.0;
double compass_roll = 0.0;
double compass_pitch = 0.0;
long last_imu_print = 0;

// ---- Compass LEDs
int north_indicator;
int heading_indicator;
int robot_direction_indicator;
bool at_heading = false;
double tracking_heading = 320.94;
long last_diff_print = 0;
uint32_t pixel_colours[NUMPIXELS];
uint8_t pixel_states[NUMPIXELS]; // 0 = off, 1 = indicators, 2 = turning
uint32_t bg_pixel_colour = 0;
enum all_drive_states {
  IDLE_DRIVE,
  FORWARD,
  RIGHT,
  HARD_RIGHT,
  LEFT,
  HARD_LEFT,
  REVERSE
};
int DRIVE_STATE = IDLE_DRIVE;
int PREV_DRIVE_STATE = IDLE_DRIVE;
int bg_alt = 0;
long last_bg_change = 0;

// ---- Air quality
// Connects to Serial5
/* Connected to Bt on Bowie Brain Board:
 *  Gnd
 *  todo
 */
String air_quality_data = "";

// ---- DHT22
/*  Connects to Sensors Input 2
 *  Gnd - Gnd (Blue)
 *  5V - N/C
 *  3.3V - Vin (Red)
 *  54 - N/C
 *  55 - N/C
 *  56 - Vout (Yellow)
 *  57 - N/C
 */
dht DHT;
#define DHT22_PIN 56
long last_dht22 = 0;


// ---- NAVCAM BOWIE COMMS
void receivedAction(Msg m);
void commsTimeout();
void controllerAdded();
void controllerRemoved();

long last_latency_print = 0;
long last_update = 0;
boolean flip_coin = false;
int sensor_send_index = 0;
long navpi_last_rx = 0;
bool blink456 = false;

Msg random_periodic1 = navcam_comms.msg_none;
Msg random_periodic2 = navcam_comms.msg_none;


// ---- OTHER
long current_time = 0;
bool completed_action[] = { false, false, false, false, false, false };
void control(Msg m);

// ---- CLAW
#define SERVO_CLAW 24
#define SERVO_CLAW_KEY 7
BowieClaw bowieclaw = BowieClaw();
void clawInterrupt(int key, int val);

//  ---
// claw variables
int frame_delay = 500;
bool FRAME_CLAW_DEBUG = true;
int frame_duration = 0;
int step_delay = 0;
int step_bound = 0;
int step_size = 0;
int total_time = 0;
int end_pos = 0;
int end_start = 0;
int end_final = 0;
int end_slice = 0;
int claw_pos = 0;
int claw_start = 0;
int claw_final = 0;
int claw_slice = 0;
int arm_pos = 0;
int arm_start = 0;
int arm_final = 0;
int arm_slice = 0;
int lid_pos = 0;
int lid_start = 0;
int lid_final = 0;
int lid_slice = 0;
boolean claw_open = true;

// pi capture blinky
bool do_the_blink = false;
long last_blinky = 0;
int blinky_step = 0;


void setup() {
  Serial.begin(9600);
  Serial5.begin(9600); // uRADMonitor Air Quality sensor
  Serial5.setTimeout(100);
  Serial4.begin(9600); // BT module on Pixy port
  Serial1.begin(9600); // Pi

  bowie = MegaBowieShoreline();
  
  // GPS
  gpsInit();

  // LEDs
  ledsInit();

  // Compass
  compassInit();
  
  // Bowie
  bowie.setStates(); // step 1) init the default states
  //bowie.enableTransparentMode(); // step 2) change settings xbee pro series 3b in transparent mode
  bowie.begin(); // step 3) begin
  bowie.setRobotID(ROBOT_ID);
  bowie.disableDefaultActions();
  bowie.set_control_callback(control);

  // Claw
  bowieclaw.begin();
  bowieclaw.setServoClawPin(SERVO_CLAW);
  bowieclaw.setServoInterruptCallback(clawInterrupt);
  bowieclaw.initServo();
  
  // Bowie Comms Navcam Pi (PI_CONN = Serial1 = Pi port)
  navcam_comms.begin();
  navcam_comms.setCommLed(COMM_LED);
  navcam_comms.set_received_action_callback(receivedAction);
  navcam_comms.set_comms_timeout_callback(commsTimeout);
  navcam_comms.set_controller_added_callback(controllerAdded);
  navcam_comms.set_controller_removed_callback(controllerRemoved);
  navcam_comms.disableReply();
  navcam_comms.initComms(PI_CONN, 9600);

  // Yellow Bowie custom adjustments
  /*
  bowie.bowiearm.SERVO_OFFSET = 50;
  bowie.bowiehopper.SERVO_HOPPER_OFFSET = 0;

  bowie.disableDefaultActions();
  bowie.disableSimpleMessageForwarding(); // disabling the forwarding of simple messages to the arduino port (which in this case is the bluetooth module for the virtual interface)
  bowie.disableMessageForwarding(); // same as above, but for the API messages. could be fun to try this enabled, and see what the latency is like

  bowie.bowiedrive.MOTOR_DIR_FWD = true;
  bowie.bowiedrive.MOTOR_DIR_REV = false;
  bowie.bowiedrive.TURN_DIR = false;
  */

  // Battery
  pinMode(A19, INPUT);

  // Hello world!
  Serial4 << "Bowie has initialised" << endl;

  bowie.bowiehopper.unparkLid();

}

void loop() {
  
  current_time = millis();

  float batt_val = ((float)analogRead(A19)/1024.0*3.3*2.76);

  if(CONSOLE_DEBUG) Serial << "Updating bowie..." << endl;
  bowie.update(true); // true = force no sleep - necessary for webvr

  // Lights on by default
  //bowie.bowielights.setLight(99, MIN_BRIGHTNESS);

  // Navcam Pi
  comms_navpi();
  
  // GPS
  if(CONSOLE_DEBUG) Serial << "Updating GPS..." << endl;
  gpsUpdate();

  // Compass
  if(CONSOLE_DEBUG) Serial << "Updating compass..." << endl;
  compassUpdate();

  if(CONSOLE_DEBUG) Serial << "Updating leds..." << endl;
  compassIndicatorLeds();
  compassDriveLeds();
  if(current_time-last_bg_change >= 100) {
    if(bg_alt == 0) {
      bg_alt = 1;
    } else if(bg_alt == 1) {
      bg_alt = 0;
    }
    last_bg_change = current_time;
  }
  updateLeds();
  
  // Updating our logging sensors
  if(CONSOLE_DEBUG) Serial << "Updating logging sensors..." << endl;
  bowie.bowielogger.setLogData_f(LOG_IMU_PITCH, compass_pitch);
  bowie.bowielogger.setLogData_f(LOG_IMU_ROLL, compass_roll);
  bowie.bowielogger.setLogData_f(LOG_IMU_YAW, 0.0);
  bowie.bowielogger.setLogData_f(LOG_COMPASS_HEADING, compass_heading);
  bowie.bowielogger.setLogData_u8(LOG_GPS_SATS, gps.satellites.value());
  bowie.bowielogger.setLogData_f(LOG_GPS_HDOP, gps.hdop.value());
  bowie.bowielogger.setLogData_f(LOG_GPS_LATITUDE, gps.location.lat());
  bowie.bowielogger.setLogData_f(LOG_GPS_LONGITUDE, gps.location.lng());
  bowie.bowielogger.setLogData_f(LOG_GPS_ALTITUDE, gps.altitude.meters());
  bowie.bowielogger.setLogData_u16(LOG_BATTERY_SENSOR, batt_val); // 1.33 for single cell, 2.76 for both
  
  // ----

  // Updating the DHT-22
  if(CONSOLE_DEBUG) Serial << "Updating DHT-22..." << endl;
  if(millis()-last_dht22 >= 1000) {
    // update it
    int chk = DHT.read22(DHT22_PIN);
    bowie.bowielogger.setLogData_f(LOG_HUMIDITY, DHT.humidity);
    bowie.bowielogger.setLogData_f(LOG_TEMPERATURE, DHT.temperature);
    if(CONSOLE_DEBUG) {
      Serial << "DHT22---> " << DHT.humidity << "% humidity " << DHT.temperature << "deg C" << endl;
      Serial << "AQ---> " << air_quality_data << endl;
      Serial << "Batt val---> " << batt_val << endl;
    }

    // -- just for a demo
    loadingIndicator(80, 20, 100, 180);
    int r = (int)random(0,2);
    bool le_dir = false;
    if(r == 0) le_dir = true;
    bowie.bowiedrive.goSpeed(le_dir, 255, 80);
    bowie.bowiedrive.goSpeed(false, 0, 20);
    unloadingIndicator(80, 0, 0, 0);
    // --
    
    last_dht22 = millis();
  }

  // Send it to our bluetooth console
  if(CONSOLE_DEBUG) Serial << "Sending to BT..." << endl;
  if(current_time-last_bowie_log >= 1000) {
    if(CONSOLE_DEBUG) {
      Serial4 << "Bowie Log ---> " << DHT.humidity << " % humidity " << DHT.temperature << " deg C";
      Serial4 << " @ (";
      Serial4.print(gps.location.lat(), 8);
      Serial4 << ", ";
      Serial4.print(gps.location.lng(), 8);
      Serial4 << ") heading = " << compass_heading << " deg" << " batt = " << batt_val << endl;
    }
    last_bowie_log = current_time;
  }
  
  // Read serial console for simple test commands
  if(CONSOLE_DEBUG) Serial << "Reading serial console..." << endl;
  while(Serial.available() > 0) {
    char c = Serial.read();
    if(c == 'e') {
      bowie.emptyScoop();
    }
  }
  
}

bool isNoneMsg(Msg m) {
  if(m.action != bowie.bowiecomms_xbee.msg_none.action) return false;
  if(m.pck1.cmd != bowie.bowiecomms_xbee.msg_none.pck1.cmd) return false;
  if(m.pck1.key != bowie.bowiecomms_xbee.msg_none.pck1.key) return false;
  if(m.pck1.val != bowie.bowiecomms_xbee.msg_none.pck1.val) return false;
  if(m.pck2.cmd != bowie.bowiecomms_xbee.msg_none.pck2.cmd) return false;
  if(m.pck2.key != bowie.bowiecomms_xbee.msg_none.pck2.key) return false;
  if(m.pck2.val != bowie.bowiecomms_xbee.msg_none.pck2.val) return false;
  if(m.delim != bowie.bowiecomms_xbee.msg_none.delim) return false;
  return true;
}

void control(Msg m) {
  // Received an action from the controller. The data is
  // packed into the Msg struct. Core actions with this data
  // will be done inside of the main robot mission program.
  // All Msgs will be passed through to this function, even
  // if there is / is not a core action associated with it.
  // You can do custom actions with this data here.

  if(isNoneMsg(m)) {
    Serial << "RX msg_none" << endl;
    return;
  }

  Serial << "---XBEE RECEIVED ACTION---" << endl;
  Serial << "action: " << m.action << endl;
  Serial << "command: " << m.pck1.cmd << endl;
  Serial << "key: " << m.pck1.key << endl;
  Serial << "val: " << m.pck1.val << endl;
  Serial << "command: " << m.pck2.cmd << endl;
  Serial << "key: " << m.pck2.key << endl;
  Serial << "val: " << m.pck2.val << endl;
  Serial << "delim: " << m.delim << endl;

  Packet packets[2] = { m.pck1, m.pck2 };

  bowie.bowielogger.logEvent(now(), m.action, m.pck1.cmd, m.pck1.key, m.pck1.val, m.pck2.cmd, m.pck2.key, m.pck2.val, m.delim);

  bowie.joystickDriveControl(m);
  bowie.joystickArmControl(m);

  // 1. control
  controlMode(m, packets);
  
  // 2. precision  mode
  precisionMode(m, packets);

  // 3. all the commands for claw mode
  clawMode(m, packets);

  // (not in use) 2. precision arm mode
  // precisionArmMode();

}


void controlMode(Msg m, Packet packets[]) {

  // ---- CONTROL MODE ----
  if(m.action == '#') {

    if(packets[0].key == 1) { // control

      if(packets[0].cmd == 'P' && packets[0].val == 1) { // drive
        // handled on op interface side
      }

      if(packets[0].cmd == 'Y' && packets[0].val == 1) { // arm
        // handled on op interface side
      }

      if(packets[0].cmd == 'G' && packets[0].val == 1) { // collect

        if(bowie.performing_large_action) return; // don't do this action if we're already doing one
        bowie.performing_large_action = true;
        
        clawOpen();
        clawLower();
        bowie.bowiedrive.goSpeed(false, 255, 250);
        bowie.bowiedrive.goSpeed(false, 0, 0);
        clawClamp(3);
        clawPick();

        bowie.performing_large_action = false;
        
      }

      if(packets[0].cmd == 'W' && packets[0].val == 1) { // open - close

        if(claw_open) {
          clawClose();
        } else {
          clawOpen();  
        }
        claw_open = !claw_open;
        
      }

      if(packets[0].cmd == 'B' && packets[0].val == 1) { // deposit

        // -- dump
        if(bowie.performing_large_action) return; // don't do this action if we're already doing one
        bowie.performing_large_action = true;
        
        moveTheArm(ARM_HOME, 1, 2);
        arm_final = ARM_HOME;
        clawClamp(3);
        clawPick();
        
        //bowie.deposit();
        
        bowie.performing_large_action = false;
        
      }

      if(packets[0].cmd == 'N' && packets[0].val == 1) { // wave

        // -- wave
        // unpark
        bowie.bowiearm.unparkArm();
        bowie.bowiescoop.unparkEnd();

        if(!claw_open) clawOpen();
        wave();
        clawClose();
        claw_open = false;

        // park
        bowie.bowiescoop.parkEnd();
        
      }

    }
    
  }
  // ----

}



void precisionMode(Msg m, Packet packets[]) {

  if(m.action == '#') {

  // ---- PRECISION MODE ----
    if(packets[0].key == 2) { // precision mode

      if(packets[0].cmd == 'P' && packets[0].val == 1) { // L
        bowie.bowiedrive.goSpeed(false, true, 255, 0);
        bowie.bowiedrive.goSpeed(true, false, 255, 0);
        delay(500);
        bowie.bowiedrive.goSpeed(true, 0, 0);
      }

      if(packets[0].cmd == 'Y' && packets[0].val == 1) { // Fwd
        bowie.bowiedrive.goSpeed(false, 255, 250);
        bowie.bowiedrive.goSpeed(false, 0, 0);
      }

      if(packets[0].cmd == 'G' && packets[0].val == 1) { // R
        bowie.bowiedrive.goSpeed(true, true, 255, 0);
        bowie.bowiedrive.goSpeed(false, false, 255, 0);
        delay(500);
        bowie.bowiedrive.goSpeed(true, 0, 0);
      }

      if(packets[0].cmd == 'W' && packets[0].val == 1) { // Arm +
        // go up slightly
        int arm_pos = bowie.bowiearm.getArmPos();
        int new_pos = arm_pos+100;
        if(new_pos > ARM_MAX) new_pos = ARM_MAX;
        moveTheArm(new_pos, 3, 3);
      }

      if(packets[0].cmd == 'B' && packets[0].val == 1) { // Bwd
        bowie.bowiedrive.goSpeed(true, 255, 250);
        bowie.bowiedrive.goSpeed(true, 0, 0);
      }

      if(packets[0].cmd == 'N' && packets[0].val == 1) { // Arm -
        // go down slightly
        int arm_pos = bowie.bowiearm.getArmPos();
        int new_pos = arm_pos-100;
        if(new_pos < ARM_MIN) new_pos = ARM_MIN;
        moveTheArm(new_pos, 3, 3);
      }
      
    }
    // -----

  }

}

void clawMode(Msg m, Packet packets[]) {

  if(m.action == '#') {

  // ---- CLAW MODE ----
    if(packets[0].key == 3) {

      if(packets[0].cmd == 'P') { // 1. Lower
        if(packets[0].val == 1 && completed_action[0] == false) { 
          
          // claw lower
          clawLower();
          
          resetCompletes();
          completed_action[0] = true;
        } else if(packets[0].val == 0) {
          completed_action[0] = false;
        }
      }

      if(packets[0].cmd == 'Y') { // 3. Hold
        if(packets[0].val == 1 && completed_action[1] == false) { 
          
          // claw pick
          clawPick();
          
          resetCompletes();
          completed_action[1] = true;
        } else if(packets[0].val == 0) {
          completed_action[1] = false;
        }
      }

      if(packets[0].cmd == 'G') { // 4. Deposit
        if(packets[0].val == 1 && completed_action[2] == false) { 
          
          // -- dump
          if(bowie.performing_large_action) return; // don't do this action if we're already doing one
          bowie.performing_large_action = true;
          
          moveTheArm(ARM_HOME, 1, 2);
          arm_final = ARM_HOME;
          clawClamp(3);
          clawPick();
          
          //bowie.deposit();
          
          bowie.performing_large_action = false;

          resetCompletes();
          completed_action[2] = true;
        } else if(packets[0].val == 0) {
          completed_action[2] = false;
        }
      }

      if(packets[0].cmd == 'W') { // 2. Small
        if(packets[0].val == 1 && completed_action[5] == false) { 
          
          // claw clamp small
          clawClamp(3);
          
          resetCompletes();
          completed_action[5] = true;
        } else if(packets[0].val == 0) {
          completed_action[5] = false;
        }
      }

      if(packets[0].cmd == 'B') { // 2. Medium
        if(packets[0].val == 1 && completed_action[4] == false) { 
          
          // claw clamp medium
          clawClamp(1);
          
          resetCompletes();
          completed_action[4] = true;
        } else if(packets[0].val == 0) {
          completed_action[4] = false;
        }
      }

      if(packets[0].cmd == 'N') { // 2. Large
        if(packets[0].val == 1 && completed_action[3] == false) { 
          
          // claw clamp large
          clawClamp(0);
          
          resetCompletes();
          completed_action[3] = true;
        } else if(packets[0].val == 0) {
          completed_action[3] = false;
        }
      }
      
    }
    // --------

  }

}






void precisionDriveMode(Msg m, Packet packets[]) {

  if(m.action == '#') {

  // ---- PRECISION DRIVE MODE ----
    if(packets[0].key == 2) { // drive

      if(packets[0].cmd == 'P' && packets[0].val == 1) { // +
        bowie.bowiedrive.goSpeed(true, 255, 250);
        bowie.bowiedrive.goSpeed(true, 0, 0);
      }

      if(packets[0].cmd == 'Y' && packets[0].val == 1) { // -
        bowie.bowiedrive.goSpeed(false, 255, 250);
        bowie.bowiedrive.goSpeed(true, 0, 0);
      }

      if(packets[0].cmd == 'G' && packets[0].val == 1) { // L
        bowie.bowiedrive.goSpeed(false, true, 255, 0);
        bowie.bowiedrive.goSpeed(true, false, 255, 0);
        delay(500);
        bowie.bowiedrive.goSpeed(true, 0, 0);
      }

      if(packets[0].cmd == 'W' && packets[0].val == 1) { // R
        bowie.bowiedrive.goSpeed(true, true, 255, 0);
        bowie.bowiedrive.goSpeed(false, false, 255, 0);
        delay(500);
        bowie.bowiedrive.goSpeed(true, 0, 0);
      }

      if(packets[0].cmd == 'B' && packets[0].val == 1) { // med
        // handled on op interface side
      }

      if(packets[0].cmd == 'N' && packets[0].val == 1) { // slow
        // handled on op interface side
      }
      
    }
    // -----

  }

}


void precisionArmMode(Msg m, Packet packets[]) {

  if(m.action == '#') {

  // ---- PRECISION ARM MODE ----
    if(packets[0].key == 2) { // arm

      if(packets[0].cmd == 'P' && packets[0].val == 1) { // +
        // go up slightly
        int arm_pos = bowie.bowiearm.getArmPos();
        int new_pos = arm_pos+100;
        if(new_pos > ARM_MAX) new_pos = ARM_MAX;
        moveTheArm(new_pos, 3, 3);
      }

      if(packets[0].cmd == 'Y' && packets[0].val == 1) { // -
        // go down slightly
        int arm_pos = bowie.bowiearm.getArmPos();
        int new_pos = arm_pos-100;
        if(new_pos < ARM_MIN) new_pos = ARM_MIN;
        moveTheArm(new_pos, 3, 3);
      }

      if(packets[0].cmd == 'G' && packets[0].val == 1) { // tilt end down
        
        bowie.bowiescoop.moveEnd(END_MAX-200, 2, 1);
        /*
        if(bowie.bowiearm.getArmParked()) {
          bowie.bowiearm.unparkArm();
        } else {
          bowie.bowiearm.parkArm();
        }
        if(bowie.bowiescoop.getEndParked()) {
          bowie.bowiescoop.unparkEnd();
        } else {
          bowie.bowiescoop.parkEnd();
        }
        */
      }

      if(packets[0].cmd == 'W' && packets[0].val == 1) { // up
        if(packets[0].val == 1 && completed_action[3] == false) {
          moveTheArm(ARM_MAX, 3, 3);
          resetCompletes();
          completed_action[3] = true;
        } else if(packets[0].val == 0) {
          completed_action[3] = false;
        }
      }

      if(packets[0].cmd == 'B' && packets[0].val == 1) { // mid
        if(packets[0].val == 1 && completed_action[4] == false) {
          moveTheArm(ARM_HOME, 3, 3);
          resetCompletes();
          completed_action[4] = true;
        } else if(packets[0].val == 0) {
          completed_action[4] = false;
        }
      }

      if(packets[0].cmd == 'N' && packets[0].val == 1) { // down
        if(packets[0].val == 1 && completed_action[5] == false) {
          moveTheArm(ARM_MIN, 3, 3);
          resetCompletes();
          completed_action[5] = true;
        } else if(packets[0].val == 0) {
          completed_action[5] = false;
        }
      }
      
    }
    // ----

  }

}


void clawInterrupt(int key, int val) {
  // This will be called any time there is a long
  // duration movement - in between the delays.
  // This is a good place to check for serial
  // messages. Val = the position.
  
  if(key == SERVO_CLAW_KEY) {
    // It's the scoop!
  }
}
